#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "usart.h"
#include "lcd.h"
#include "DS18B20.h"

void processConnections(void);

//gwna strona
const char WebPage[] PROGMEM = "HTTP/1.1 200 OK\r\n"
"Connection: Close\r\n\r\n"
"<html><head><title>Witaj</title></head><body>"
"Temperatura: %05.2f&deg;C<BR>"
"<form method='post'>"
"<input type='submit' name='action' value='ON'>"
"<input type='submit' name='action' value='OFF'>"
"</form></body></html>";

//brak strony
const char NotFound[] PROGMEM = "HTTP/1.1 404 Not Found\r\n"
"Connection: Close\r\n\r\n"
"<html><head><title>404 Not Found</title>"
"</head><body><h1>Nic tu nie ma</h1>"
"</body></html>";

//przekierowanie na gwn
const char Redirect[] PROGMEM = "HTTP/1.1 302 Found\r\n"
"Location: /\r\n\r\n";

char buffer[512];
char IPaddress[16];
uint16_t bufferIndex = 0;
FILE usartStream;
volatile uint8_t ok = 0;
volatile uint8_t gotIP = 0;

enum RequestType {
	GET_ROOT,
	GET_OTHER,
	POST_DATA
};
	
struct {
	uint8_t httpRequestReceived;
	enum RequestType url;
	uint8_t postAction;
} connections[4] = {0};

void waitForOK(void) {
	while(!ok);
	ok = 0;
}

void waitForIP(void) {
	while(!gotIP);
	gotIP = 0;
}

int main(void){
	DDRD |= _BV(DDD2) | _BV(DDD3);
	//reset moduu
	PORTD &= ~_BV(PD2);
	_delay_ms(10);
	PORTD |= _BV(PD2);
	lcdInit();
	lcdInitPrintf();
	usartInit(115200);
	sei();
	UCSRB |= _BV(RXCIE);
	fdev_setup_stream(&usartStream, &usartPut, &usartGet, _FDEV_SETUP_RW);
	//czekaj na poczenie z sieci i uzyskanie IP
	waitForIP();
	//pobierz i wywietl IP
	fprintf(&usartStream, "AT+CIFSR\r\n");
	waitForOK();
	printf(IPaddress);
	fprintf(&usartStream, "AT+CIPMUX=1\r\n");
	waitForOK();
	fprintf(&usartStream, "AT+CIPSERVER=1,80\r\n");
	waitForOK();
    while(1) processConnections();
}

void processConnections(void) {
	static float temperature = 0;
	for (uint8_t conn = 0; conn < 5; conn++) {
		if (connections[conn].httpRequestReceived) {
			connections[conn].httpRequestReceived = 0;
			if (connections[conn].url == GET_ROOT) {
				//odelij gwn stron
				fprintf(&usartStream, "AT+CIPSEND=%d,%d\r\n", conn, strlen_P(WebPage) - 1);
				waitForOK();
				fprintf_P(&usartStream, WebPage, temperature);
			}
			if (connections[conn].url == GET_OTHER) {
				//odelij 404
				fprintf(&usartStream, "AT+CIPSEND=%d,%d\r\n", conn, strlen_P(NotFound));
				waitForOK();
				fprintf_P(&usartStream, NotFound);
			}
			if (connections[conn].url == POST_DATA) {
				//przekieruj na gwn stron
				fprintf(&usartStream, "AT+CIPSEND=%d,%d\r\n", conn, strlen_P(Redirect));
				waitForOK();
				fprintf_P(&usartStream, Redirect);
				//przecz LEDa
				if (connections[conn].postAction) {
					PORTD |= _BV(PD3);
				} else {
					PORTD &= ~_BV(PD3);
				}
			}
			waitForOK();
			//zamknij poczenie
			fprintf(&usartStream, "AT+CIPCLOSE=%d\r\n", conn);
			waitForOK();
		}
		//pobierz aktualn temperatur z DS18B20
		getTemperature(&temperature);
	}
}

ISR(USART_RXC_vect) {
	static uint16_t size = 0;
	static uint8_t ipd = 0;
	static uint8_t conn = 0;
	uint8_t value = UDR;
	if (bufferIndex < sizeof(buffer)) {
		buffer[bufferIndex++] = value;
	}
	if (ipd) {//jeli odbieramy dane z przegldarki
		if (size == 0) {
			if (value == ':') {//pobierz numer poczenia i liczb bajtw
				size = atoi(buffer + 7);
				conn = atoi(buffer + 5);
				bufferIndex = 0;
			}
		} else {
			if (bufferIndex == size) {
				//odebrano wszystkie bajty
				//sprawd rodzaj dania
				connections[conn].httpRequestReceived = 1;
				if (strncmp(buffer, "GET ", 4) == 0) {
					if (strncmp(buffer, "GET / ", 6) == 0) {
						connections[conn].url = GET_ROOT;
					} else {
						connections[conn].url = GET_OTHER;
					}
				}
				if (strncmp(buffer, "POST ", 5) == 0) {
					connections[conn].url = POST_DATA;
					//jeli POST, odnajd parametr action
					char * postData = strstr(buffer, "\r\n\r\n");
					if (postData) {
						if (strncmp(postData + 4, "action=ON", 9) == 0) {
							connections[conn].postAction = 1;
						} else {
							connections[conn].postAction = 0;
						}
					}
				}
				size = 0;
				ipd = 0;
				bufferIndex = 0;
			}
		}
	} else {
		if (value == '\n') {
			buffer[bufferIndex] = 0;
			bufferIndex = 0;
			//sprawd status z moduu
			if (strncmp(buffer, "OK", 2) == 0) ok = 1;
			if (strncmp(buffer, "SEND OK", 7) == 0) ok = 1;
			if (strncmp(buffer, "WIFI GOT IP", 11) == 0) gotIP = 1;
			if (strncmp(buffer, "+CIFSR:STAIP", 12) == 0) {
				//jeli modu zwrci IP, skopiuj je do IPaddress
				char * ip = strstr(buffer, "\"") + 1;
				for (uint8_t i = 0; i < 15; i++) {
					if (*ip != '\"') {
						IPaddress[i] = *ip++;
					} else {
						IPaddress[i] = 0;
						break;
					}
					IPaddress[15] = 0;
				}
			}
		}
		if (bufferIndex == 5) {
			if (strncmp(buffer, "+IPD,", 5) == 0) ipd = 1;
		}
	}
}
